"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ViewportList = void 0;
var jsx_runtime_1 = require("react/jsx-runtime");
var react_1 = require("react");
var MIN_INDEX = 0;
var IS_OVERFLOW_ANCHOR_SUPPORTED = (function () {
    try {
        return window.CSS.supports('overflow-anchor: auto');
    }
    catch (error) {
        return false;
    }
})();
var PROP_NAME_FOR_Y_AXIS = {
    top: 'top',
    bottom: 'bottom',
    clientHeight: 'clientHeight',
    scrollHeight: 'scrollHeight',
    scrollTop: 'scrollTop',
    overflowY: 'overflowY',
    height: 'height',
    minHeight: 'minHeight',
    maxHeight: 'maxHeight'
};
var PROP_NAME_FOR_X_AXIS = {
    top: 'left',
    bottom: 'right',
    clientHeight: 'clientWidth',
    scrollHeight: 'scrollWidth',
    scrollTop: 'scrollLeft',
    overflowY: 'overflowX',
    height: 'width',
    minHeight: 'minWidth',
    maxHeight: 'maxWidth'
};
var getStyle = function (propName, size) {
    var _a;
    return (_a = {},
        _a[propName.minHeight] = size,
        _a[propName.height] = size,
        _a[propName.maxHeight] = size,
        _a.overflowAnchor = 'none',
        _a.pointerEvents = 'none',
        _a.userSelect = 'none',
        _a.padding = 0,
        _a.margin = 0,
        _a.border = 'none',
        _a);
};
var normalizeValue = function (min, value, max) { return Math.max(Math.min(value, max), min); };
var useIsomorphicLayoutEffect = typeof window !== 'undefined' ? react_1.useLayoutEffect : react_1.useEffect;
var useRequestAnimationFrame = function (func) {
    var stepRef = (0, react_1.useRef)(func);
    stepRef.current = func;
    (0, react_1.useEffect)(function () {
        var frameId;
        var frame = function () {
            frameId = requestAnimationFrame(frame);
            stepRef.current();
        };
        frame();
        return function () {
            cancelAnimationFrame(frameId);
        };
    }, []);
};
var ViewportListInner = function (_a, ref) {
    var viewportRef = _a.viewportRef, _b = _a.items, items = _b === void 0 ? [] : _b, itemMinSize = _a.itemMinSize, _c = _a.margin, margin = _c === void 0 ? 0 : _c, _d = _a.overscan, overscan = _d === void 0 ? 1 : _d, _e = _a.axis, axis = _e === void 0 ? 'y' : _e, _f = _a.initialIndex, initialIndex = _f === void 0 ? 0 : _f, _g = _a.initialAlignToTop, initialAlignToTop = _g === void 0 ? true : _g, _h = _a.initialOffset, initialOffset = _h === void 0 ? 0 : _h, _j = _a.fixed, fixed = _j === void 0 ? false : _j, children = _a.children;
    var propName = axis === 'y' ? PROP_NAME_FOR_Y_AXIS : PROP_NAME_FOR_X_AXIS;
    var maxIndex = items.length - 1;
    var itemMinSizeWithMargin = itemMinSize + margin;
    var overscanSize = overscan * itemMinSizeWithMargin;
    var _k = (0, react_1.useState)(function () {
        var normalizedInitialIndex = normalizeValue(MIN_INDEX, initialIndex, maxIndex);
        return [normalizedInitialIndex, normalizedInitialIndex];
    }), _l = _k[0], startIndex = _l[0], endIndex = _l[1], setIndexes = _k[1];
    var normalizedStartIndex = normalizeValue(MIN_INDEX, startIndex, maxIndex);
    var normalizedEndIndex = normalizeValue(normalizedStartIndex, endIndex, maxIndex);
    var topRef = (0, react_1.useRef)(null);
    var bottomRef = (0, react_1.useRef)(null);
    var cacheRef = (0, react_1.useRef)([]);
    var scrollToIndexRef = (0, react_1.useRef)(startIndex ? { index: startIndex, alignToTop: initialAlignToTop, offset: initialOffset } : null);
    var scrollCompensationEndIndexRef = (0, react_1.useRef)(-1);
    var topStyle = (0, react_1.useMemo)(function () {
        return getStyle(propName, cacheRef.current
            .slice(MIN_INDEX, normalizedStartIndex)
            .reduce(function (sum, next) { return sum + (next - itemMinSize); }, normalizedStartIndex * itemMinSizeWithMargin));
    }, [itemMinSize, itemMinSizeWithMargin, normalizedStartIndex, propName]);
    var bottomStyle = (0, react_1.useMemo)(function () {
        return getStyle(propName, cacheRef.current
            .slice(normalizedEndIndex + 1, maxIndex)
            .reduce(function (sum, next) { return sum + (next - itemMinSize); }, itemMinSizeWithMargin * (maxIndex - normalizedEndIndex)));
    }, [propName, normalizedEndIndex, maxIndex, itemMinSizeWithMargin, itemMinSize]);
    useRequestAnimationFrame(function () {
        if (!topRef.current ||
            !bottomRef.current ||
            !topRef.current.nextSibling ||
            !bottomRef.current.previousSibling) {
            return;
        }
        var viewport = (viewportRef && viewportRef.current) || document.documentElement;
        var viewportRect = viewport.getBoundingClientRect();
        var topLimit = normalizeValue(0, viewportRect[propName.top], document.documentElement[propName.clientHeight]) -
            overscanSize;
        var bottomLimit = normalizeValue(0, viewportRect[propName.bottom], document.documentElement[propName.clientHeight]) +
            overscanSize;
        // eslint-disable-next-line @typescript-eslint/ban-ts-comment
        // @ts-ignore
        var topElementRect = topRef.current.nextSibling.getBoundingClientRect();
        // eslint-disable-next-line @typescript-eslint/ban-ts-comment
        // @ts-ignore
        var bottomElementRect = bottomRef.current.previousSibling.getBoundingClientRect();
        var maxItemsCountInViewPort = Math.ceil((bottomLimit - topLimit) / itemMinSizeWithMargin);
        var nextStartIndex = startIndex;
        var nextEndIndex = endIndex;
        // If scroll needed
        if (scrollToIndexRef.current) {
            var targetIndex = normalizeValue(MIN_INDEX, scrollToIndexRef.current.index, maxIndex);
            // If our target scroll index is rendered
            if (targetIndex >= startIndex && targetIndex <= endIndex) {
                var index = startIndex;
                var element = topRef.current.nextSibling;
                while (element && element !== bottomRef.current) {
                    if (index === targetIndex) {
                        // eslint-disable-next-line @typescript-eslint/ban-ts-comment
                        // @ts-ignore
                        element.scrollIntoView(scrollToIndexRef.current.alignToTop);
                        if (scrollToIndexRef.current.offset &&
                            Math.abs(viewport[propName.scrollHeight] - viewport[propName.clientHeight]) < 1) {
                            viewport[propName.scrollTop] += scrollToIndexRef.current.offset;
                        }
                        scrollToIndexRef.current = null;
                        break;
                    }
                    element = element.nextSibling;
                    ++index;
                }
                return;
            }
            nextStartIndex = targetIndex - maxItemsCountInViewPort;
            nextEndIndex = targetIndex + maxItemsCountInViewPort;
            // If each item height is fixed
        }
        else if (fixed) {
            nextStartIndex = Math.floor((topLimit - topRef.current.getBoundingClientRect()[propName.top]) / itemMinSizeWithMargin);
            nextEndIndex = nextStartIndex + maxItemsCountInViewPort;
            // If first rendered item is below bottom
        }
        else if (topElementRect[propName.top] >= bottomLimit) {
            var diff = topElementRect[propName.top] - bottomLimit;
            nextEndIndex = startIndex;
            while (diff >= 0 && nextEndIndex > MIN_INDEX) {
                diff -= (cacheRef.current[--nextEndIndex] || itemMinSize) + margin;
            }
            scrollCompensationEndIndexRef.current = startIndex;
            nextStartIndex = nextEndIndex - maxItemsCountInViewPort;
            // If last rendered item is above top
        }
        else if (bottomElementRect[propName.bottom] + margin <= topLimit) {
            var diff = topLimit - bottomElementRect[propName.bottom] + margin;
            nextStartIndex = endIndex;
            while (diff >= 0 && nextStartIndex < maxIndex) {
                diff -= (cacheRef.current[++nextStartIndex] || itemMinSize) + margin;
            }
            nextEndIndex = nextStartIndex + maxItemsCountInViewPort;
        }
        else {
            // If first rendered item is above top
            if (topElementRect[propName.bottom] + margin < topLimit) {
                ++nextStartIndex;
                // If first rendered item is below top
            }
            else if (topElementRect[propName.top] >= topLimit) {
                var diff = topElementRect[propName.top] - topLimit;
                while (diff >= 0 && nextStartIndex > MIN_INDEX) {
                    diff -= (cacheRef.current[--nextStartIndex] || itemMinSize) + margin;
                }
                scrollCompensationEndIndexRef.current = startIndex;
            }
            // If last rendered item is above bottom
            if (bottomElementRect[propName.bottom] + margin <= bottomLimit) {
                var diff = bottomLimit - bottomElementRect[propName.bottom] - margin;
                while (diff >= 0 && nextEndIndex < maxIndex) {
                    diff -= (cacheRef.current[++nextEndIndex] || itemMinSize) + margin;
                }
                // If last rendered item is below bottom
            }
            else if (bottomElementRect[propName.top] > bottomLimit) {
                --nextEndIndex;
            }
        }
        nextStartIndex = normalizeValue(MIN_INDEX, nextStartIndex, maxIndex);
        nextEndIndex = normalizeValue(nextStartIndex, nextEndIndex, maxIndex);
        if (nextStartIndex !== startIndex || nextEndIndex !== endIndex) {
            if (fixed) {
                cacheRef.current = [];
            }
            else {
                var index = startIndex;
                var element = topRef.current.nextSibling;
                while (element && index < nextStartIndex && element !== bottomRef.current) {
                    // eslint-disable-next-line @typescript-eslint/ban-ts-comment
                    // @ts-ignore
                    cacheRef.current[index++] = element[propName.clientHeight];
                    element = element.nextSibling;
                }
                index = endIndex;
                element = bottomRef.current.previousSibling;
                while (element && index > nextEndIndex && element !== topRef.current) {
                    // eslint-disable-next-line @typescript-eslint/ban-ts-comment
                    // @ts-ignore
                    cacheRef.current[index++] = element[propName.clientHeight];
                    element = element.previousSibling;
                }
            }
            setIndexes([nextStartIndex, nextEndIndex]);
        }
    });
    useIsomorphicLayoutEffect(function () {
        if (scrollCompensationEndIndexRef.current === -1) {
            return;
        }
        if (!topRef.current || !topRef.current.nextSibling || IS_OVERFLOW_ANCHOR_SUPPORTED) {
            scrollCompensationEndIndexRef.current = -1;
            return;
        }
        var viewport = (viewportRef && viewportRef.current) || document.documentElement;
        var index = startIndex;
        var element = topRef.current.nextSibling;
        var sizeDiff = 0;
        while (element && index < scrollCompensationEndIndexRef.current && element !== bottomRef.current) {
            // eslint-disable-next-line @typescript-eslint/ban-ts-comment
            // @ts-ignore
            sizeDiff += element[propName.clientHeight] - (cacheRef.current[index] || itemMinSize);
            element = element.nextSibling;
            ++index;
        }
        if (sizeDiff && Math.abs(viewport[propName.scrollHeight] - viewport[propName.clientHeight]) < 1) {
            viewport.style[propName.overflowY] = 'hidden';
            viewport[propName.scrollTop] += sizeDiff;
            viewport.style[propName.overflowY] = '';
        }
        scrollCompensationEndIndexRef.current = -1;
        // eslint-disable-next-line react-hooks/exhaustive-deps
    }, [startIndex]);
    (0, react_1.useImperativeHandle)(ref, function () { return ({
        scrollToIndex: function (index, alignToTop, offset) {
            if (index === void 0) { index = -1; }
            if (alignToTop === void 0) { alignToTop = true; }
            if (offset === void 0) { offset = 0; }
            scrollToIndexRef.current = { index: index, alignToTop: alignToTop, offset: offset };
        }
    }); }, []);
    return ((0, jsx_runtime_1.jsxs)(react_1.Fragment, { children: [(0, jsx_runtime_1.jsx)("div", { ref: topRef, style: topStyle }, void 0), items.slice(startIndex, endIndex + 1).map(function (item, index) { return children(item, startIndex + index, items); }), (0, jsx_runtime_1.jsx)("div", { ref: bottomRef, style: bottomStyle }, void 0)] }, void 0));
};
exports.ViewportList = (0, react_1.forwardRef)(ViewportListInner);
// eslint-disable-next-line import/no-default-export
exports.default = exports.ViewportList;
